/*
 * livepatch_bsc1178066
 *
 * Fix for CVE-2020-0433, bsc#1178066
 *
 *  Upstream commits:
 *  287922eb0b18 ("block: defer timeouts to a workqueue")
 *  71f79fb3179e ("blk-mq: Allow timeouts to run while queue is freezing")
 *  f5bbbbe4d635 ("blk-mq: sync the update nr_hw_queues with
 *                 blk_mq_queue_tag_busy_iter")
 *  530ca2c9bd69 ("blk-mq: Allow blocking queue tag iter callbacks")
 *  a9a808084d6a ("block: Remove the synchronize_rcu() call from
 *                 __blk_mq_update_nr_hw_queues()")
 *
 *  SLE12-SP2 and -SP3 commits:
 *  bc8a70a10eeab626b59b14c9a0554f3328bf30a1
 *  bf382463a755e38dc6ed9e2410590a3532a9de02
 *
 *  SLE12-SP4, SLE12-SP5 and SLE15-SP1 commits:
 *  d9357044c1638512cbb4c49ae3b162eb0a87e7d7
 *  269bf1a15ac2908fa4bd39ab21b895ab32435316
 *
 *  SLE15 commits:
 *  d9357044c1638512cbb4c49ae3b162eb0a87e7d7
 *  8d03667896c6844230a62946bcce053a08d831e1
 *  7fb1c0878daef6fedfe1936ceafb60a636a43b97
 *
 *  SLE15-SP2 commit:
 *  not affected
 *
 *
 *  Copyright (c) 2021 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* klp-ccp: from block/blk-mq-tag.c */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/blk-mq.h>

/* klp-ccp: from include/linux/blkdev.h */
static void (*klpe_blk_queue_exit)(struct request_queue *q);

/* klp-ccp: from block/blk.h */
#include <linux/idr.h>
#include <linux/blk-mq.h>
/* klp-ccp: from block/blk-stat.h */
#include <linux/kernel.h>
#include <linux/blkdev.h>
#include <linux/ktime.h>
#include <linux/rcupdate.h>
#include <linux/timer.h>

/* klp-ccp: from block/blk-mq-tag.h */
struct blk_mq_tags {
	unsigned int nr_tags;
	unsigned int nr_reserved_tags;

	atomic_t active_queues;

	struct sbitmap_queue bitmap_tags;
	struct sbitmap_queue breserved_tags;

	struct request **rqs;
	struct request **static_rqs;
	struct list_head page_list;
};

static inline bool blk_mq_hw_queue_mapped(struct blk_mq_hw_ctx *hctx)
{
	return hctx->nr_ctx && hctx->tags;
}

/* klp-ccp: from block/blk-mq-tag.c */
struct bt_iter_data {
	struct blk_mq_hw_ctx *hctx;
	busy_iter_fn *fn;
	void *data;
	bool reserved;
};

static bool bt_iter(struct sbitmap *bitmap, unsigned int bitnr, void *data)
{
	struct bt_iter_data *iter_data = data;
	struct blk_mq_hw_ctx *hctx = iter_data->hctx;
	struct blk_mq_tags *tags = hctx->tags;
	bool reserved = iter_data->reserved;
	struct request *rq;

	if (!reserved)
		bitnr += tags->nr_reserved_tags;
	rq = tags->rqs[bitnr];

	/*
	 * We can hit rq == NULL here, because the tagging functions
	 * test and set the bit before assining ->rqs[].
	 */
	if (rq && rq->q == hctx->queue)
		iter_data->fn(hctx, rq, iter_data->data, reserved);
	return true;
}

static void bt_for_each(struct blk_mq_hw_ctx *hctx, struct sbitmap_queue *bt,
			busy_iter_fn *fn, void *data, bool reserved)
{
	struct bt_iter_data iter_data = {
		.hctx = hctx,
		.fn = fn,
		.data = data,
		.reserved = reserved,
	};

	sbitmap_for_each_set(&bt->sb, bt_iter, &iter_data);
}

void klpp_blk_mq_queue_tag_busy_iter(struct request_queue *q, busy_iter_fn *fn,
		void *priv)
{
	struct blk_mq_hw_ctx *hctx;
	int i;

	/*
	 * __blk_mq_update_nr_hw_queues will update the nr_hw_queues and
	 * queue_hw_ctx after freeze the queue. So we could use q_usage_counter
	 * to avoid race with it. __blk_mq_update_nr_hw_queues will users
	 * synchronize_rcu to ensure all of the users go out of the critical
	 * section below and see zeroed q_usage_counter.
	 */
	/*
	 * Fix CVE-2020-0433
	 *  -5 lines, +2 lines
	 */
	if (!percpu_ref_tryget(&q->q_usage_counter))
		return;

	queue_for_each_hw_ctx(q, hctx, i) {
		struct blk_mq_tags *tags = hctx->tags;

		/*
		 * If not software queues are currently mapped to this
		 * hardware queue, there's nothing to check
		 */
		if (!blk_mq_hw_queue_mapped(hctx))
			continue;

		if (tags->nr_reserved_tags)
			bt_for_each(hctx, &tags->breserved_tags, fn, priv, true);
		bt_for_each(hctx, &tags->bitmap_tags, fn, priv, false);
	}
	/*
	 * Fix CVE-2020-0433
	 *  -1 line, +1 line
	 */
	(*klpe_blk_queue_exit)(q);
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1178066.h"
#include "../kallsyms_relocs.h"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "blk_queue_exit", (void *)&klpe_blk_queue_exit},
};

int livepatch_bsc1178066_init(void)
{
	return __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
}
